@@ -1,7 +1,9 @@
 /*
  * Copyright (C) 2007, 2008 OpenedHand Ltd.
+ * Copyright (C) 2012 Intel Corporation.
  *
  * Authors: Jorn Baayen <jorn@openedhand.com>
+ *          Jens Georg <jensg@openismus.com>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -32,6 +34,7 @@
 #include "gupnp-didl-lite-object.h"
 #include "gupnp-didl-lite-object-private.h"
 #include "gupnp-didl-lite-descriptor-private.h"
+#include "gupnp-didl-lite-writer-private.h"
 
 #include "xml-util.h"
 
@@ -107,10 +110,10 @@ static int
 compare_node_name (const char *a, const char *b)
 {
         const char *p;
-        int len;
+        int len, result;
 
         if (a[0] == '@')
-                /* Filer is for top-level property */
+                /* Filter is for top-level property */
                 return -1;
 
         p = strstr (a, "@");
@@ -120,7 +123,15 @@ compare_node_name (const char *a, const char *b)
         else
                 len = strlen (a);
 
-        return strncmp (a, b, len);
+        result = strncmp (a, b, len);
+
+        if (result == 0) {
+            /* Avoid that we return a match although only prefixes match like
+             * in upnp:album and upnp:albumArtUri, cf. bgo#687462 */
+            return strlen (b) - len;
+        }
+
+        return result;
 }
 
 static gboolean
@@ -198,7 +209,8 @@ filter_attributes (xmlNode             *node,
 static void
 filter_node (xmlNode             *node,
              GList               *allowed,
-             GUPnPDIDLLiteWriter *writer)
+             GUPnPDIDLLiteWriter *writer,
+             gboolean             tags_only)
 {
         xmlNode *child;
         GList   *forbidden = NULL;
@@ -206,7 +218,8 @@ filter_node (xmlNode             *node,
         gboolean is_container;
         const char *container_class = NULL;
 
-        filter_attributes (node, allowed, writer);
+        if (!tags_only)
+                filter_attributes (node, allowed, writer);
 
         if (strcmp ((const char *) node->name, "container") == 0) {
             is_container = TRUE;
@@ -250,9 +263,41 @@ filter_node (xmlNode             *node,
         /* Recurse */
         for (child = node->children; child != NULL; child = child->next)
                 if (!xmlNodeIsText (child))
-                        filter_node (child, allowed, writer);
+                        filter_node (child, allowed, writer, tags_only);
 }
 
+static void
+apply_filter (GUPnPDIDLLiteWriter *writer,
+              const char          *filter,
+              gboolean             tags_only)
+{
+        char **tokens;
+        GList *allowed = NULL;
+        unsigned short i;
+        xmlNode *node;
+
+        g_return_if_fail (GUPNP_IS_DIDL_LITE_WRITER (writer));
+        g_return_if_fail (filter != NULL);
+
+        if (filter[0] == '*')
+                return;         /* Wildcard */
+
+        tokens = g_strsplit (filter, ",", -1);
+        g_return_if_fail (tokens != NULL);
+
+        for (i = 0; tokens[i] != NULL; i++)
+                allowed = g_list_append (allowed, tokens[i]);
+
+        for (node = writer->priv->xml_node->children;
+             node != NULL;
+             node = node->next)
+                filter_node (node, allowed, writer, tags_only);
+
+        g_list_free (allowed);
+        g_strfreev (tokens);
+}
+
+
 static void
 gupnp_didl_lite_writer_init (GUPnPDIDLLiteWriter *writer)
 {
@@ -482,6 +527,44 @@ gupnp_didl_lite_writer_add_item (GUPnPDIDLLiteWriter *writer)
         return GUPNP_DIDL_LITE_ITEM (object);
 }
 
+/**
+ * gupnp_didl_lite_writer_add_container_child_item:
+ * @writer: #GUPnPDIDLLiteWriter
+ * @container: #GUPnPDIDLLiteContainer
+ *
+ * Add a child item to a container. This is only useful in DIDL_S playlist
+ * creation.
+ *
+ * Returns: (transfer full): A new #GUPnPDIDLLiteItem object. Unref after
+ * usage.
+ **/
+GUPnPDIDLLiteItem *
+gupnp_didl_lite_writer_add_container_child_item
+                                        (GUPnPDIDLLiteWriter    *writer,
+                                         GUPnPDIDLLiteContainer *container)
+{
+        xmlNode *item_node, *container_node;
+        GUPnPDIDLLiteObject *object;
+        GUPnPXMLDoc *doc;
+
+        g_return_val_if_fail (GUPNP_IS_DIDL_LITE_CONTAINER (container), NULL);
+
+        object = GUPNP_DIDL_LITE_OBJECT (container);
+        container_node = gupnp_didl_lite_object_get_xml_node (object);
+
+        item_node = xmlNewChild (container_node,
+                                 NULL,
+                                 (xmlChar *) "item",
+                                 NULL);
+
+        object = gupnp_didl_lite_object_new_from_xml (item_node,
+                                                      writer->priv->xml_doc,
+                                                      writer->priv->upnp_ns,
+                                                      writer->priv->dc_ns,
+                                                      writer->priv->dlna_ns);
+        return GUPNP_DIDL_LITE_ITEM (object);
+}
+
 /**
  * gupnp_didl_lite_writer_add_container:
  * @writer: A #GUPnPDIDLLiteWriter
@@ -613,28 +696,28 @@ void
 gupnp_didl_lite_writer_filter (GUPnPDIDLLiteWriter *writer,
                                const char          *filter)
 {
-        char **tokens;
-        GList *allowed = NULL;
-        unsigned short i;
-        xmlNode *node;
+        apply_filter (writer, filter, FALSE);
+}
 
-        g_return_if_fail (GUPNP_IS_DIDL_LITE_WRITER (writer));
-        g_return_if_fail (filter != NULL);
-
-        if (filter[0] == '*')
-                return;         /* Wildcard */
-
-        tokens = g_strsplit (filter, ",", -1);
-        g_return_if_fail (tokens != NULL);
-
-        for (i = 0; tokens[i] != NULL; i++)
-                allowed = g_list_append (allowed, tokens[i]);
-
-        for (node = writer->priv->xml_node->children;
-             node != NULL;
-             node = node->next)
-                filter_node (node, allowed, writer);
-
-        g_list_free (allowed);
-        g_strfreev (tokens);
+/**
+ * gupnp_didl_lite_writer_filter_tags:
+ * @writer: A #GUPnPDIDLLiteWriter
+ * @filter: A filter string
+ *
+ * Clears the DIDL-Lite XML document of the properties not specified in the
+ * @filter. The passed filter string would typically come from the 'Filter'
+ * argument of Browse or Search actions from a ContentDirectory control point.
+ * Please refer to Section 2.3.15 of UPnP AV ContentDirectory version 3
+ * specification for details on this string.
+ *
+ * In contrast to gupnp_didl_lite_writer_filter(), this function only removes
+ * unwanted tags but leaves all attributes in-place.
+ *
+ * Return value: None.
+ **/
+void
+gupnp_didl_lite_writer_filter_tags (GUPnPDIDLLiteWriter *writer,
+                                    const char          *filter)
+{
+        apply_filter (writer, filter, TRUE);
 }
